from Contact import Contact 
from Atom import Atom
import re

ContactError = 'contact error'

class ContactList(list):
    """Class ContactList

    The ContactList class works as a container for instances of the class
    Contacts.

    The ContactList class also has a number of attributes and methods for
    summing up the number of differnet types of contacts stored in the list.
    The different types of contacts are:
    1) polar_contacts (direct polar contacts between atoms of the subAtom
                       types MacroMoleculeAtom or LigandAtom)
    2) apolar_contacts (as above but apolar instead of polar)
    3) polar_water_contacts (polar contacts between WaterProbes and
                             MacroMoleculeAtoms or LigandAtoms)
    4) apolar_water_contacts (as above but apolar instead of polar)
    5) polar_probe_contacts (VoidProbe contacts to a polar MacroMoleculeAtom
                             or a LigandAtom)
    6) apolarProbeContacts (as above but to apolar atoms)
    7) polar_surface_water_contacts (polar contacts between surface
                                     WaterProbes and MacroMoleculeAtoms or
                                     LigandAtoms)
    8) apolar_surface_water_contacts (as above but apolar)
    9) polar_cleft_water_contacts (polar contacts between clefte WaterProbes
                                    and MacroMoleculeAtoms or LigandAtoms)
    10) apolar_cleft_water_contacts (as above but apolar)
    11) polar_buried_water_contacts (polar contacts between buried WaterProbes
                                    and MacroMoleculeAtoms or LigandAtoms)
    12) apolar_buried_water_contacts (as above but apolar)
    13) polar_further_water_contacts (polar contacts between further
                                     WaterProbes and MacroMoleculeAtoms or
                                     LigandAtoms)
    14) apolar_buried_water_contacts (as above but apolar)
    15) complimentary_contacts (apolar-apolar or polar-polar atoms form the
                                contact)
    16) not_complimentary_contacts (polar-apolar or apolar-polar atoms form
                                    the contact)

    """
    
    # Operations
    def __add__(self, other):
        """function __add__
        
        This is a bit of a hack for accounting purposes
        
        """
        tmp = ContactList()
        tmp._total_contacts = ( self.total_contacts
            + other.total_contacts )
        tmp._polar_contacts = ( self.polar_contacts
            + other.polar_contacts )
        tmp._apolar_contacts = ( self.apolar_contacts
            + other.apolar_contacts )
        tmp._polar_water_contacts = ( self.polar_water_contacts
            + other.polar_water_contacts )
        tmp._apolar_water_contacts = ( self.apolar_water_contacts
            + other.apolar_water_contacts )
        tmp._polar_probe_contacts = ( self.polar_probe_contacts
            + other.polar_probe_contacts )
        tmp._apolar_probe_contacts = ( self.apolar_probe_contacts
            + other.apolar_probe_contacts )
        tmp._polar_surface_water_contacts = (
            self.polar_surface_water_contacts
            + other.polar_surface_water_contacts )
        tmp._apolar_surface_water_contacts = (
            self.apolar_surface_water_contacts
            + other.apolar_surface_water_contacts )
        tmp._polar_cleft_water_contacts = (
            self.polar_cleft_water_contacts
            + other.polar_cleft_water_contacts )
        tmp._apolar_cleft_water_contacts = (
            self.apolar_cleft_water_contacts
            + other.apolar_cleft_water_contacts )
        tmp._polar_buried_water_contacts = (
            self.polar_buried_water_contacts
            + other.polar_buried_water_contacts )
        tmp._apolar_buried_water_contacts = (
            self.apolar_buried_water_contacts
            + other.apolar_buried_water_contacts )
        tmp._polar_further_water_contacts = (
            self.polar_further_water_contacts
            + other.polar_further_water_contacts )
        tmp._apolar_further_water_contacts = (
            self.apolar_further_water_contacts
            + other.apolar_further_water_contacts )
        tmp._complimentary_contacts = (
            self.complimentary_contacts
            + other.complimentary_contacts )
        tmp._not_complimentary_contacts = (
            self.not_complimentary_contacts
            + other.not_complimentary_contacts )
        return tmp

    def __sub__(self, other):
        """function __sub__
        
        This is a bit of a hack for accounting purposes
        
        """
        tmp = ContactList()
        tmp._total_contacts = ( self.total_contacts
            - other.total_contacts )
        tmp._polar_contacts = ( self.polar_contacts
            - other.polar_contacts )
        tmp._apolar_contacts = ( self.apolar_contacts
            - other.apolar_contacts )
        tmp._polar_water_contacts = ( self.polar_water_contacts
            - other.polar_water_contacts )
        tmp._apolar_water_contacts = ( self.apolar_water_contacts
            - other.apolar_water_contacts )
        tmp._polar_probe_contacts = ( self.polar_probe_contacts
            - other.polar_probe_contacts )
        tmp._apolar_probe_contacts = ( self.apolar_probe_contacts
            - other.apolar_probe_contacts )
        tmp._polar_surface_water_contacts = (
            self.polar_surface_water_contacts
            - other.polar_surface_water_contacts )
        tmp._apolar_surface_water_contacts = (
            self.apolar_surface_water_contacts
            - other.apolar_surface_water_contacts )
        tmp._polar_cleft_water_contacts = (
            self.polar_cleft_water_contacts
            - other.polar_cleft_water_contacts )
        tmp._apolar_cleft_water_contacts = (
            self.apolar_cleft_water_contacts
            - other.apolar_cleft_water_contacts )
        tmp._polar_buried_water_contacts = (
            self.polar_buried_water_contacts
            - other.polar_buried_water_contacts )
        tmp._apolar_buried_water_contacts = (
            self.apolar_buried_water_contacts
            - other.apolar_buried_water_contacts )
        tmp._polar_further_water_contacts = (
            self.polar_further_water_contacts
            - other.polar_further_water_contacts )
        tmp._apolar_further_water_contacts = (
            self.apolar_further_water_contacts
            - other.apolar_further_water_contacts )
        tmp._complimentary_contacts = (
            self.complimentary_contacts
            - other.complimentary_contacts )
        tmp._not_complimentary_contacts = (
            self.not_complimentary_contacts
            - other.not_complimentary_contacts )
        return tmp

    def __div__(self, other):
        """function __div__
        
        This is a bit of a hack for accounting purposes
        
        """
        tmp = ContactList()
        # Make sure that at least one parameter is a float
        denominator = float(other)
        tmp._total_contacts = ( self.total_contacts
            / denominator )
        tmp._polar_contacts = ( self.polar_contacts
            / denominator )
        tmp._apolar_contacts = ( self.apolar_contacts
            / denominator )
        tmp._polar_water_contacts = ( self.polar_water_contacts
            / denominator )
        tmp._apolar_water_contacts = ( self.apolar_water_contacts
            / denominator )
        tmp._polar_probe_contacts = ( self.polar_probe_contacts
            / denominator )
        tmp._apolar_probe_contacts = ( self.apolar_probe_contacts
            / denominator )
        tmp._polar_surface_water_contacts = (
            self.polar_surface_water_contacts
            / denominator )
        tmp._apolar_surface_water_contacts = (
            self.apolar_surface_water_contacts
            / denominator )
        tmp._polar_cleft_water_contacts = (
            self.polar_cleft_water_contacts
            / denominator )
        tmp._apolar_cleft_water_contacts = (
            self.apolar_cleft_water_contacts
            / denominator )
        tmp._polar_buried_water_contacts = (
            self.polar_buried_water_contacts
            / denominator )
        tmp._apolar_buried_water_contacts = (
            self.apolar_buried_water_contacts
            / denominator )
        tmp._polar_further_water_contacts = (
            self.polar_further_water_contacts
            / denominator )
        tmp._apolar_further_water_contacts = (
            self.apolar_further_water_contacts
            / denominator )
        tmp._complimentary_contacts = (
            self.complimentary_contacts
            / denominator )
        tmp._not_complimentary_contacts = (
            self.not_complimentary_contacts
            / denominator )
        return tmp

    def contact_input(self, atom1, atom2, polarity):
        """function contact_input
        
        atom1: Atom
        atom2: Atom
        polarity: int

        This method is used to append contacts to the list.
        It is used by the Assembly.setContacts() method, which also determines
        the intiger value of the polarity argument. The polarity argument has
        the can have the following values:
        1) Polar  -- (0)
        2) Apolar -- (2)
        
        """
        self.append(Contact(atom1, atom2, polarity))
    
    @property
    def total_contacts(self):
        """
        Return total number of contacts.
        """
        return self._total_contacts
    
    @property
    def polar_contacts(self):
        """
        Return number of polar contact.
        """
        return self._polar_contacts
    
    @property
    def apolar_contacts(self):
        """
        Return number of apolar contacts.
        """
        return self._apolar_contacts
    
    @property
    def polar_water_contacts(self):
        """
        Return number of polar water contacts.
        """
        return self._polar_water_contacts
    
    @property
    def apolar_water_contacts(self):
        """
        Return number of apolar water contacts.
        """
        return self._apolar_water_contacts
    
    @property
    def polar_probe_contacts(self):
        """
        Return number of polar probe contacts.
        """
        return self._polar_probe_contacts
    
    @property
    def apolar_probe_contacts(self):
        """
        Return number of apolar probe contacts.
        """
        return self._apolar_probe_contacts
    
    @property
    def polar_surface_water_contacts(self):
        """
        Return number of polar surface water contacts.
        """
        return self._polar_surface_water_contacts
    
    @property
    def apolar_surface_water_contacts(self):
        """
        Return number of apolar surface water contacts.
        """
        return self._apolar_surface_water_contacts

    @property
    def polar_cleft_water_contacts(self):
        """
        Return number of polar cleft water contacts.
        """
        return self._polar_cleft_water_contacts
    
    @property
    def apolar_cleft_water_contacts(self):
        """
        Return number of apolar cleft water contacts.
        """
        return self._apolar_cleft_water_contacts

    @property
    def polar_buried_water_contacts(self):
        """
        Return polar buried water contacts.
        """
        return self._polar_buried_water_contacts
    
    @property
    def apolar_buried_water_contacts(self):
        """
        Return number of apolar buried water contacts.
        """
        return self._apolar_buried_water_contacts

    @property
    def polar_further_water_contacts(self):
        """
        Return number of polar further water contacts.
        """
        return self._polar_further_water_contacts
    
    @property
    def apolar_further_water_contacts(self):
        """
        Return number of apolar further water contacts.
        """
        return self._apolar_further_water_contacts

    @property
    def complimentary_contacts(self):
        """
        Return number of complimentary contacts.
        """
        return self._complimentary_contacts
    
    @property
    def not_complimentary_contacts(self):
        """
        Return number of not complimentary contacts.
        """
        return self._not_complimentary_contacts
    
    def calculate_contact_counts(self):
        """function calculate_contact_counts
        
        This method adds up all the contacts in the contactList and stores the
        results in the appropriate attributes.

        """
        # reset counts to zero
        self._total_contacts = 0
        self._polar_contacts = 0
        self._apolar_contacts = 0
        self._polar_water_contacts = 0
        self._apolar_water_contacts = 0
        self._polar_probe_contacts = 0
        self._apolar_probe_contacts = 0
        self._apolar_surface_water_contacts = 0
        self._polar_surface_water_contacts = 0
        self._apolar_cleft_water_contacts = 0
        self._polar_cleft_water_contacts = 0
        self._apolar_buried_water_contacts = 0
        self._polar_buried_water_contacts = 0
        self._apolar_further_water_contacts = 0
        self._polar_further_water_contacts = 0
        self._complimentary_contacts = 0
        self._not_complimentary_contacts = 0
        # loop through contacts in contacList
        contacts = self
        for contact in contacts:
            atom1 = contact._atom1
            atom2 = contact._atom2
            # sanity checking
            if atom1._sub_atom_type == 3:
                if atom1.experimental:
                    continue
                raise ContactError, "atom 1 water! should be macroMol. or lig."
            if atom1._sub_atom_type == 4:
                raise ContactError, "atom 1 probe! should be macroMol. or lig."
            # counting
            if atom2._sub_atom_type == 1 or atom2._sub_atom_type == 2:
                # Add up complimentartiy
                if contact.contact_complimentarity == 1:
                    self._complimentary_contacts = (
                        self.complimentary_contacts + 1 )
                else:
                    self._not_complimentary_contacts = ( 
                        self.not_complimentary_contacts + 1 )
                # Add up contacts
                if contact._contact_polarity == 0:
                    self._polar_contacts = self.polar_contacts + 1
                    self._total_contacts = self.total_contacts + 1
                elif contact._contact_polarity == 2:
                    self._apolar_contacts = self.apolar_contacts + 1
                    self._total_contacts = self.total_contacts + 1
                else:
                    raise ContactError, "contactPolarity not (0) or (2)"
            elif atom2._sub_atom_type == 3:
                if atom2.water_type == 1:
                    if contact._contact_polarity == 0:
                        self._polar_surface_water_contacts= (
                            self.polar_surface_water_contacts + 1 )
                    elif contact._contact_polarity == 2:
                        self._apolar_surface_water_contacts= (
                            self.apolar_surface_water_contacts + 1 )
                    else:
                        raise ContactError, "contactPolarity not (0) or (2)"
                if atom2.water_type > 1:
                    if contact._contact_polarity == 0:
                        self._polar_cleft_water_contacts= (
                            self.polar_cleft_water_contacts + 1 )
                    elif contact._contact_polarity == 2:
                        self._apolar_cleft_water_contacts= (
                            self.apolar_cleft_water_contacts + 1 )
                    else:
                        raise ContactError, "contactPolarity not (0) or (2)"
                if atom2.water_type == 0:
                    if contact._contact_polarity == 0:
                        self._polar_buried_water_contacts= (
                            self.polar_buried_water_contacts + 1 )
                    elif contact._contact_polarity == 2:
                        self._apolar_buried_water_contacts= (
                            self.apolar_buried_water_contacts + 1 )
                    else:
                        raise ContactError, "contactPolarity not (0) or (2)"
                if atom2.water_type < 0:
                    if contact._contact_polarity == 0:
                        self._polar_further_water_contacts= (
                            self.polar_further_water_contacts + 1 )
                    elif contact._contact_polarity == 2:
                        self._apolar_further_water_contacts= (
                            self.apolar_further_water_contacts + 1 )
                    else:
                        raise ContactError, "contactPolarity not (0) or (2)"
                # Set total water count
                if contact._contact_polarity == 0:
                    self._polar_water_contacts = (
                        self.polar_water_contacts + 1 )
                elif contact._contact_polarity == 2:
                    self._apolar_water_contacts = (
                        self.apolar_water_contacts + 1 )
                else:
                    raise ContactError, "contactPolarity not (0) or (2)"
            if atom2._sub_atom_type == 4:
                # Only count probe contacts to probes with a radius greater
                # than or equal to 0.8 Angstroms
                if atom2.probe_radius >= 0.8:
                    if contact._contact_polarity == 0:
                        self._polar_probe_contacts = self.polar_probe_contacts + 1
                    elif contact._contact_polarity == 2:
                        self._apolar_probe_contacts = self.apolar_probe_contacts + 1
                    else:
                        raise ContactError, "contactPolarity not (0) or (2)"

    def contacts_string(self):
        """
        Return contact string.
        """
        s = """Number of total contacts               : %.1f
               Number of polar contacts               : %.1f
               Number of apolar contacts              : %.1f
               Number of polar water contacts         : %.1f
               Number of apolar water contacts        : %.1f
               Number of polar surface water contacts : %.1f
               Number of apolar surface water contacts: %.1f
               Number of polar cleft water contacts   : %.1f
               Number of apolar cleft water contacts  : %.1f
               Number of polar buried water contacts  : %.1f
               Number of apolar buried water contacts : %.1f
               Number of polar further water contacts : %.1f
               Number of apolar further water contacts: %.1f
               Number of polar cavity contacts        : %.1f
               Number of apolar cavity contacts       : %.1f""" % (
               float(self.total_contacts),
               float(self.polar_contacts),
               float(self.apolar_contacts),
               float(self.polar_water_contacts),
               float(self.apolar_water_contacts),
               float(self.polar_surface_water_contacts),
               float(self.apolar_surface_water_contacts),
               float(self.polar_cleft_water_contacts),
               float(self.apolar_cleft_water_contacts),
               float(self.polar_buried_water_contacts),
               float(self.apolar_buried_water_contacts),
               float(self.polar_further_water_contacts),
               float(self.apolar_further_water_contacts),
               float(self.polar_probe_contacts),
               float(self.apolar_probe_contacts) )    
        return re.sub(r'\n\s+', '\n', s)

    def contacts_csv(self):
        """function contacts_csv
        
        This method returns the contacts as a CSV string for printing.

        returns string
        """
        s = []
        s.append("%.1f," % float(self.total_contacts))
        s.append("%.1f," % float(self.polar_contacts))
        s.append("%.1f," % float(self.apolar_contacts))
        s.append("%.1f," % float(self.polar_water_contacts))
        s.append("%.1f," % float(self.apolar_water_contacts))
        s.append("%.1f," % float(self.polar_surface_water_contacts))
        s.append("%.1f," % float(self.apolar_surface_water_contacts))
        s.append("%.1f," % float(self.polar_cleft_water_contacts))
        s.append("%.1f," % float(self.apolar_cleft_water_contacts))
        s.append("%.1f," % float(self.polar_buried_water_contacts))
        s.append("%.1f," % float(self.apolar_buried_water_contacts))
        s.append("%.1f," % float(self.polar_further_water_contacts))
        s.append("%.1f," % float(self.apolar_further_water_contacts))
        s.append("%.1f," % float(self.polar_probe_contacts))
        s.append("%.1f" % float(self.apolar_probe_contacts))
        return ''.join(s)

    def clear(self):
        """
        Remove atom references and empty list
        """
        for contact in self:
            contact.clear()
        self.__delslice__(0, self.__len__())
        
# Test the contactList class 'python ContactList.py'
if __name__ == "__main__":
    import doctest
    import unittest
    suite = doctest.DocFileSuite('doc_tests/test_ContactList.txt',
        optionflags=doctest.ELLIPSIS + doctest.NORMALIZE_WHITESPACE)
    unittest.TextTestRunner().run(suite)
