from VoidProbe import VoidProbe

class Void(list):
    """Class Void

    The class Void is a container for VoidProbes, just like MacroMolecule and
    Ligand are containers for MacroMoleculeAtoms and LigandAtoms.

    >>> from Void import Void
    >>> V = Void()
    >>> V.atom_input(1, 'PROBE', 'PROBE', 0.0, 0.0, 0.0, 4, 0)
    >>> V.atom_input(1, 'PROBE', 'PROBE', 2.0, 0.0, 0.0, 4, 0)
    >>> len(V)
    2
    >>> for void in V:
    ...     print '%.2f' % void.x
    0.00
    2.00
    
    """

    def atom_input(self, atomOriginalId, atomOriginalName, atomType, x, y, z,
        subAtomType, moleculeInstance):
        """function atom_input
        
        atomOriginalId: int
        atomOriginalName: string
        atomType: string
        x: float
        y: float
        z: float
        subAtomType: int
        moleculeInstance: int

        This method is used to input a new VoidProbe into the list.

        See the Atom class for explanations of the arguments passed into this
        method.
        
        """
        self.append(VoidProbe(atomOriginalId,
                              atomOriginalName,
                              atomType,
                              x, y, z,
                              subAtomType,
                              moleculeInstance))
    
    def clear(self):
        """
        Clear list.

        Including neighbor lists.

        """
        for atom in self:
            atom.clear_neighbors()
        del self[:]
        
# Test Void class by 'python Hydration.py'
if __name__ == "__main__":
    import doctest
    doctest.testmod()
    v = Void()
